// (c) flatz

#include "common.h"
#include "util.h"

#include "aes.h"
#include "sha1.h"
#include "mtrand.h"

#define ARCTOOL_VERSION "0.1.0"

#define ARCHIVE_HEADER_SIZE_GENERIC 0x40
#define ARCHIVE_HEADER_SIZE_PER_CONSOLE 0x30
#define ARCHIVE_TYPE_3 0x03
#define ARCHIVE_TYPE_5 0x05

#define ARCHIVE_HEADER_SIZE 0x40
#define ARCHIVE_KEY_BLOCK_SIZE 0x40
#define ARCHIVE_KEY_BLOCK_KEY_OFFSET 0x2C
#define ARCHIVE_DATA_BLOCK_SIZE 0x10000

#define ARCHIVE_CONSOLE_ID_SIZE 0x10
#define ARCHIVE_KEY_SIZE 0x14
#define ARCHIVE_DATA_KEY_SIZE 0x10
#define ARCHIVE_DATA_IV_SIZE 0x10
#define ARCHIVE_MAC_KEY_SIZE 0x14
#define ARCHIVE_HASH_SIZE 0x14

static int s_cmd_encrypt = 0;
static int s_cmd_decrypt = 0;

char* g_p_in_file_path = NULL;
char* g_p_out_file_path = NULL;
char* g_p_console_id_path = NULL;

int g_use_per_console = 0;

uint8_t* g_p_console_id = NULL;

FILE* g_p_in_file = NULL;
uint64_t g_in_file_size = 0;
FILE* g_p_out_file = NULL;

aes_context_t g_p_crypto_context = NULL;
sha1_context g_hash_context;

const uint32_t g_random_seed[] = {
	0x979BB357UL, 0x78337340UL, 0xB9286FBFUL, 0x3C4AA7E2UL,
	0x05E5564BUL, 0xF9B4DE3DUL, 0x12B73C76UL, 0xEC828F2EUL,
	0x789B15B8UL, 0x3366C0BCUL, 0xBB58195AUL, 0x2CA09E10UL,
	0x6698E648UL, 0xD8D828FCUL, 0x890350A1UL, 0xBCC0C0BCUL,
	0x2747989CUL, 0xE616FBB5UL, 0x7E15250BUL, 0xF9EA2E23UL,
	0x4AE6097DUL, 0x270738FAUL, 0x0631B677UL, 0x81600F0EUL,
	0x9E4BA357UL, 0x0B0A7BD5UL, 0xBA11E99CUL, 0xEAC0F832UL,
	0x79A14BBCUL, 0x78F6A610UL, 0x890FDB60UL, 0x156EE98EUL,
};

#if FZ_COMPILER_MSVC
	#pragma pack(push, 1)
#endif

typedef union archive_s {
	struct {
		uint8_t header_size;
		uint8_t _pad0[3];
		uint32_t type;
		uint8_t mac[ARCHIVE_HASH_SIZE];
		uint8_t seed[ARCHIVE_KEY_SIZE];
		uint8_t _pad1[16];
	};
	uint8_t raw[ARCHIVE_HEADER_SIZE];
} archive_t;

typedef union archive_key_block_s {
	struct {
		uint8_t data_key[16];
		uint8_t data_iv[16];
		uint8_t _pad0[12];
		uint8_t mac_key[ARCHIVE_MAC_KEY_SIZE];
	};
	uint8_t raw[ARCHIVE_KEY_BLOCK_SIZE];
} archive_key_block_t;

#if FZ_COMPILER_MSVC
	#pragma pack(pop)
#endif

static void show_version(void) {
	printf("arctool " ARCTOOL_VERSION " (c) 2013 by flatz\n");
	printf("\n");
}

static void show_usage(void) {
	show_version();

	printf("USAGE: arctool [options] command\n");
	printf("\n");
	printf("COMMANDS           Parameters          Explanation\n");
	printf(" -h, --help                            Print this help\n");
	printf(" -e, --encrypt     <in> <out>          Encrypt an archive file\n");
	printf(" -d, --decrypt     <in> <out>          Decrypt an archive file\n");
	printf("\n");
	printf("OPTIONS            Parameters          Explanation\n");
	printf(" -i, --cid         console-id          Console ID\n");
	printf(" -p, --per-console                     Use per-console keys\n");
}

static int parse_args(int argc, char* argv[]) {
	int num_left;
	int num_commands;
	int i;

	num_commands = 0;

	for (i = 1; i < argc; ++i) {
		num_left = argc - i;
		if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0) {
			return 0;
		} else if (strcmp(argv[i], "--encrypt") == 0 || strcmp(argv[i], "-e") == 0 && num_left >= 2) {
			g_p_in_file_path = argv[i + 1];
			g_p_out_file_path = argv[i + 2];
			s_cmd_encrypt = 1;
			num_commands++;
			i += 2;
		} else if (strcmp(argv[i], "--decrypt") == 0 || strcmp(argv[i], "-d") == 0 && num_left >= 2) {
			g_p_in_file_path = argv[i + 1];
			g_p_out_file_path = argv[i + 2];
			s_cmd_decrypt = 1;
			num_commands++;
			i += 2;
		} else if (strcmp(argv[i], "--cid") == 0 || strcmp(argv[i], "-i") == 0 && num_left >= 1) {
			g_p_console_id_path = argv[i + 1];
			i += 1;
		} else if (strcmp(argv[i], "--per-console") == 0 || strcmp(argv[i], "-p") == 0) {
			g_use_per_console = 1;
		}
	}

	if (g_p_console_id_path) {
		g_p_console_id = (uint8_t*)malloc(ARCHIVE_CONSOLE_ID_SIZE);
		memset(g_p_console_id, 0, ARCHIVE_CONSOLE_ID_SIZE);
		if (read_file(g_p_console_id_path, g_p_console_id, ARCHIVE_CONSOLE_ID_SIZE) < 0) {
			const char* p;
			free(g_p_console_id);
			for (p = g_p_console_id_path; *p != '\00'; ++p) {
				if (!isxdigit(*p)) {
					p = NULL;
					break;
				}
			}
			g_p_console_id = p != NULL ? x_to_u8_buffer(g_p_console_id_path) : NULL;
		}
	}

	return num_commands;
}

int setup_random_generator(uint32_t perturbator, const uint32_t* p_seed, int32_t count) {
	uint32_t* p_new_seed;
	int32_t i;
	
	if (p_seed && count > 0) {
		p_new_seed = (uint32_t*)malloc(count * sizeof(uint32_t));
		if (!p_new_seed)
			return -1;
		
		for (i = 0; i < count; ++i)
			p_new_seed[i] ^= perturbator;
		
		mtrand_init_by_array(p_new_seed, count);
		free(p_new_seed);
	} else {
		mtrand_init(perturbator);
	}

	return 0;
}

uint8_t next_random_byte(void) {
	return mtrand_get_next() & 0xFFUL;
}

int32_t generate_random_bytes(uint8_t* p_data, int32_t size) {
	uint32_t rnd;
	int32_t i;
	
	if (!p_data || size <= 0)
		return -1;
		
	memset(p_data, 0, size);
	for (i = 0; i < size; ++i) {
		if ((i & 3) == 0)
			rnd = mtrand_get_next();
		p_data[i] ^= (rnd >> (8 * (3 - (i & 3)))) & 0xFFUL;
	}
	
	return 0;
}


int archive_compute_keys(uint8_t header_size, uint8_t seed[ARCHIVE_KEY_SIZE], uint8_t data_key[ARCHIVE_DATA_KEY_SIZE], uint8_t data_iv[ARCHIVE_DATA_IV_SIZE], uint8_t mac_key[ARCHIVE_MAC_KEY_SIZE]) {
	aes_context_t p_context;
	archive_key_block_t key_block;
	const uint8_t* p_current_key;
	uint8_t iv[16];

	static const uint8_t generic_key[16] = {
		0xFA, 0x72, 0xCE, 0xEF, 0x59, 0xB4, 0xD2, 0x98, 0x9F, 0x11, 0x19, 0x13, 0x28, 0x7F, 0x51, 0xC7,
	};
	static const uint8_t per_console_key[16] = {
		0xEA, 0x02, 0xCE, 0xEF, 0x5B, 0xB4, 0xD2, 0x99, 0x8F, 0x61, 0x19, 0x10, 0xD7, 0x7F, 0x51, 0xC6,
	};

	if (header_size != ARCHIVE_HEADER_SIZE_GENERIC && header_size != ARCHIVE_HEADER_SIZE_PER_CONSOLE)
		return -1;
	if (header_size == ARCHIVE_HEADER_SIZE_PER_CONSOLE && !g_p_console_id)
		return -1;

	memset(key_block.raw, 0, ARCHIVE_KEY_BLOCK_SIZE);
	switch (header_size) {
		case ARCHIVE_HEADER_SIZE_GENERIC:
			memcpy(key_block.data_key, seed, ARCHIVE_KEY_SIZE);
			p_current_key = generic_key;
			break;
		case ARCHIVE_HEADER_SIZE_PER_CONSOLE:
			memcpy(key_block.data_key, g_p_console_id, ARCHIVE_CONSOLE_ID_SIZE);
			p_current_key = per_console_key;
			break;
	}

	p_context = aes_make_context();
	if (!p_context)
		return -1;

	memset(iv, 0, 16);
	aes_set_key(p_context, p_current_key, 128);
	aes_encrypt_cbc(p_context, iv, key_block.raw, ARCHIVE_KEY_BLOCK_SIZE);

	memcpy(data_key, key_block.data_key, ARCHIVE_DATA_KEY_SIZE);
	memcpy(data_iv, key_block.data_iv, ARCHIVE_DATA_IV_SIZE);
	memcpy(mac_key, key_block.mac_key, ARCHIVE_MAC_KEY_SIZE);

	aes_free_context(p_context);

	return 0;
}

int archive_encrypt(void) {
	archive_t archive;
	uint8_t data_key[ARCHIVE_DATA_KEY_SIZE];
	uint8_t data_iv[ARCHIVE_DATA_IV_SIZE];
	uint8_t mac_key[ARCHIVE_MAC_KEY_SIZE];
	uint8_t data[ARCHIVE_DATA_BLOCK_SIZE];
	uint8_t computed_mac[ARCHIVE_HASH_SIZE];
	uint8_t null_iv[AES_BLOCK_SIZE];
	uint8_t prev_block[AES_BLOCK_SIZE];
	uint64_t data_size;
	uint64_t block_size, new_block_size;
	uint64_t bytes_left;
	uint64_t num_blocks;
	uint64_t offset;
	uint32_t percent, last_percent;
	uint64_t i, j;

	printf("[*] Encrypting %s into %s...\n", g_p_in_file_path, g_p_out_file_path);

	if (g_in_file_size == 0) {
		printf("[!] Error! Invalid file size.\n");
		return -1;
	}

	memset(&archive, 0, sizeof(archive));

	archive.type = ES32(ARCHIVE_TYPE_5);
	if (g_use_per_console) {
		archive.header_size = ARCHIVE_HEADER_SIZE_PER_CONSOLE;
	} else {
		archive.header_size = ARCHIVE_HEADER_SIZE_GENERIC;
		if (generate_random_bytes(archive.seed, ARCHIVE_KEY_SIZE) < 0) {
			printf("[!] Unable to generate random sequence of bytes.\n");
			return -1;
		}
	}

	printf("[*] Computing archive keys...\n");
	if (archive_compute_keys(archive.header_size, archive.seed, data_key, data_iv, mac_key) < 0) {
		printf("[!] Error! Unable to compute keys (console ID is missing?).\n");
		return -1;
	}

	fseek(g_p_out_file, ARCHIVE_HEADER_SIZE, SEEK_SET);

	data_size = g_in_file_size;
	if (data_size > 0)
		num_blocks = 1 + (data_size - 1) / ARCHIVE_DATA_BLOCK_SIZE;
	else
		num_blocks = 0;

	printf("[*] Processing file data...\n");
	if (num_blocks != 0) {
		memset(&g_hash_context, 0, sizeof(sha1_context));

		memset(prev_block, 0, AES_BLOCK_SIZE);
		aes_set_key(g_p_crypto_context, data_key, ARCHIVE_DATA_KEY_SIZE * 8);

		memset(computed_mac, 0, ARCHIVE_HASH_SIZE);
		sha1_hmac_starts(&g_hash_context, mac_key, ARCHIVE_MAC_KEY_SIZE);

		last_percent = 0;
		for (i = 0; i < num_blocks; ++i) {
			offset = i * ARCHIVE_DATA_BLOCK_SIZE;
			block_size = i < (num_blocks - 1) ? ARCHIVE_DATA_BLOCK_SIZE : (data_size - (num_blocks - 1) * ARCHIVE_DATA_BLOCK_SIZE);
			if (_fseeki64(g_p_in_file, offset, SEEK_SET) != 0)
				return -1;
			memset(data, 0, ARCHIVE_DATA_BLOCK_SIZE);
			if (fread(data, 1, block_size, g_p_in_file) != block_size)
				return -1;
			new_block_size = (block_size / AES_BLOCK_SIZE) * AES_BLOCK_SIZE;
			bytes_left = block_size - new_block_size;
			if (new_block_size >= AES_BLOCK_SIZE) {
				aes_encrypt_cbc(g_p_crypto_context, data_iv, data, new_block_size);
				memcpy(prev_block, data + new_block_size - AES_BLOCK_SIZE, AES_BLOCK_SIZE);
			}
			if (bytes_left > 0 && bytes_left < AES_BLOCK_SIZE) {
				memset(null_iv, 0, AES_BLOCK_SIZE);
				aes_encrypt_cbc(g_p_crypto_context, null_iv, prev_block, AES_BLOCK_SIZE); 
				for (j = 0; j < bytes_left; ++j)
					data[new_block_size + j] ^= prev_block[j];
			}
			sha1_hmac_update(&g_hash_context, data, block_size);
			fwrite(data, 1, block_size, g_p_out_file);
			percent = i < (num_blocks - 1) ? (uint32_t)(i * 100 / num_blocks) : 100;
			if ((percent - last_percent) > 9 || i == num_blocks - 1) {
				printf("[*] %3d%% completed.\n", percent);
				last_percent = percent;
			}
		}

		sha1_hmac_finish(&g_hash_context, computed_mac);
		memcpy(archive.mac, computed_mac, ARCHIVE_HASH_SIZE);

		fseek(g_p_out_file, 0, SEEK_SET);
		fwrite(&archive, 1, ARCHIVE_HEADER_SIZE, g_p_out_file);
	}

	return 0;
}

int archive_decrypt(void) {
	archive_t archive;
	uint8_t data_key[ARCHIVE_DATA_KEY_SIZE];
	uint8_t data_iv[ARCHIVE_DATA_IV_SIZE];
	uint8_t mac_key[ARCHIVE_MAC_KEY_SIZE];
	uint8_t data[ARCHIVE_DATA_BLOCK_SIZE];
	uint8_t computed_mac[ARCHIVE_HASH_SIZE];
	uint8_t null_iv[AES_BLOCK_SIZE];
	uint8_t prev_block[AES_BLOCK_SIZE];
	uint64_t data_size;
	uint64_t block_size, new_block_size;
	uint64_t bytes_left;
	uint64_t num_blocks;
	uint64_t offset;
	uint32_t percent, last_percent;
	uint64_t i, j;

	printf("[*] Decrypting %s into %s...\n", g_p_in_file_path, g_p_out_file_path);

	if (g_in_file_size < ARCHIVE_HEADER_SIZE) {
		printf("[!] Error! Invalid file size.\n");
		return -1;
	}

	if (fread(&archive, 1, ARCHIVE_HEADER_SIZE, g_p_in_file) != ARCHIVE_HEADER_SIZE) {
		printf("[!] Error! Invalid file format.\n");
		return -1;
	}

	if (archive.header_size != ARCHIVE_HEADER_SIZE_GENERIC && archive.header_size != ARCHIVE_HEADER_SIZE_PER_CONSOLE) {
		printf("[!] Error! Unsupported archive type.\n");
		return -1;
	}

	if (ES32(archive.type) != ARCHIVE_TYPE_3 && ES32(archive.type) != ARCHIVE_TYPE_5) {
		printf("[!] Error! Unsupported archive type.\n");
		return -1;
	}

	printf("[*] Computing archive keys...\n");
	if (archive_compute_keys(archive.header_size, archive.seed, data_key, data_iv, mac_key) < 0) {
		printf("[!] Error! Unable to compute keys (console ID is missing?).\n");
		return -1;
	}

	data_size = g_in_file_size - ARCHIVE_HEADER_SIZE;
	if (data_size > 0)
		num_blocks = 1 + (data_size - 1) / ARCHIVE_DATA_BLOCK_SIZE;
	else
		num_blocks = 0;

	printf("[*] Processing file data...\n");
	if (num_blocks != 0) {
		memset(&g_hash_context, 0, sizeof(sha1_context));

		memset(prev_block, 0, AES_BLOCK_SIZE);
		aes_set_key(g_p_crypto_context, data_key, ARCHIVE_DATA_KEY_SIZE * 8);

		memset(computed_mac, 0, ARCHIVE_HASH_SIZE);
		sha1_hmac_starts(&g_hash_context, mac_key, ARCHIVE_MAC_KEY_SIZE);

		last_percent = 0;
		for (i = 0; i < num_blocks; ++i) {
			offset = ARCHIVE_HEADER_SIZE + i * ARCHIVE_DATA_BLOCK_SIZE;
			block_size = i < (num_blocks - 1) ? ARCHIVE_DATA_BLOCK_SIZE : (data_size - (num_blocks - 1) * ARCHIVE_DATA_BLOCK_SIZE);
			if (_fseeki64(g_p_in_file, offset, SEEK_SET) != 0)
				return -1;
			memset(data, 0, ARCHIVE_DATA_BLOCK_SIZE);
			if (fread(data, 1, block_size, g_p_in_file) != block_size)
				return -1;
			sha1_hmac_update(&g_hash_context, data, block_size);
			new_block_size = (block_size / AES_BLOCK_SIZE) * AES_BLOCK_SIZE;
			bytes_left = block_size - new_block_size;
			if (new_block_size >= AES_BLOCK_SIZE) {
				memcpy(prev_block, data + new_block_size - AES_BLOCK_SIZE, AES_BLOCK_SIZE);
				aes_decrypt_cbc(g_p_crypto_context, data_iv, data, new_block_size);
			}
			if (bytes_left > 0 && bytes_left < AES_BLOCK_SIZE) {
				memset(null_iv, 0, AES_BLOCK_SIZE);
				aes_encrypt_cbc(g_p_crypto_context, null_iv, prev_block, AES_BLOCK_SIZE); 
				for (j = 0; j < bytes_left; ++j)
					data[new_block_size + j] ^= prev_block[j];
			}
			fwrite(data, 1, block_size, g_p_out_file);
			percent = i < (num_blocks - 1) ? (uint32_t)(i * 100 / num_blocks) : 100;
			if ((percent - last_percent) > 9 || i == num_blocks - 1) {
				printf("[*] %3d%% completed.\n", percent);
				last_percent = percent;
			}
		}

		sha1_hmac_finish(&g_hash_context, computed_mac);

		if (memcmp(archive.mac, computed_mac, ARCHIVE_HASH_SIZE) != 0) {
			printf("[!] Error! Invalid HMAC.\n");
			return -1;
		}
	}

	return 0;
}

static void exit_callback(void) {
	if (g_p_out_file) {
		fclose(g_p_out_file);
		g_p_out_file = NULL;
	}

	if (g_p_in_file) {
		fclose(g_p_in_file);
		g_p_in_file = NULL;
		g_in_file_size = 0;
	}

	if (g_p_console_id) {
		free(g_p_console_id);
		g_p_console_id = NULL;
	}

	memset(&g_hash_context, 0, sizeof(sha1_context));

	if (g_p_crypto_context) {
		aes_free_context(g_p_crypto_context);
		g_p_crypto_context = NULL;
	}
}

int main(int argc, char* argv[]) {
	if (argc <= 1)
		goto error;

	atexit(&exit_callback);

	if (!parse_args(argc, argv))
		goto error;

	if (s_cmd_encrypt || s_cmd_decrypt) {
		if (setup_random_generator((uint32_t)time(NULL), g_random_seed, FZ_COUNT_OF(g_random_seed)) < 0)
			goto error;

		memset(&g_hash_context, 0, sizeof(sha1_context));
		g_p_crypto_context = aes_make_context();
		if (!g_p_crypto_context)
			goto error;

		if (get_file_size(g_p_in_file_path, &g_in_file_size) < 0)
			goto error;

		g_p_in_file = fopen(g_p_in_file_path, "rb");
		if (!g_p_in_file)
			goto error;

		g_p_out_file = fopen(g_p_out_file_path, "wb");
		if (!g_p_out_file)
			goto error;
	}

	show_version();

	if (s_cmd_encrypt)
		archive_encrypt();

	if (s_cmd_decrypt)
		archive_decrypt();

	return 0;

error:
	show_usage();
	return 1;
}