#include "util.h"
#include "platform.h"

#if FZ_PLATFORM_WIN32
	#include <io.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <polarssl/sha1.h>

uint64_t x_to_u64(const char* p_hex) {
	uint64_t result, t;
	uint32_t len;
	int32_t c;

	result = 0;
	t = 0;
	len = strlen(p_hex);

	while (len--) {
		c = *p_hex++;
		if (c >= '0' && c <= '9')
			t = c - '0';
		else if (c >= 'a' && c <= 'f')
			t = c - 'a' + 10;
		else if (c >= 'A' && c <= 'F')
			t = c - 'A' + 10;
		else
			t = 0;
		result |= t << (len * 4);
	}

	return result;
}

uint8_t* x_to_u8_buffer(const char* p_hex) {
	char tmp[3] = { 0, 0, 0 };
	uint8_t* p_result;
	uint8_t* p;
	uint32_t len;

	len = strlen(p_hex);
	if (len % 2 != 0) {
		return NULL;
	}

	p_result = (uint8_t*)malloc(len);
	memset(p_result, 0, len);
	p = p_result;

	while (len--) {
		tmp[0] = *p_hex++;
		tmp[1] = *p_hex++;
		*p++ = (uint8_t)x_to_u64(tmp);
	}

	return p_result;
}

int get_file_size(const char* p_file_path, uint64_t* p_size) {
	struct _stat64 stat_buf;
	if (!p_file_path || !p_size) {
		return -1;
	}

	if (_stat64(p_file_path, &stat_buf) < 0) {
		return -1;
	}

	*p_size = stat_buf.st_size;

	return 0;
}

int read_file(const char* p_file_path, uint8_t* p_data, uint64_t size) {
	FILE* p_file;
	struct stat stat_buf;

	if (!p_file_path || !p_data) {
		return -1;
	}

	p_file = fopen(p_file_path, "rb");
	if (!p_file) {
		return -1;
	}

	if (fstat(fileno(p_file), &stat_buf) < 0) {
		fclose(p_file);
		return -1;
	}

	if (stat_buf.st_size < size) {
		fclose(p_file);
		return -1;
	}

	memset(p_data, 0, size);

	if (fread(p_data, 1, size, p_file) != size) {
		fclose(p_file);
		return -1;
	}

	fclose(p_file);

	return 0;
}

int write_file(const char* p_file_path, uint8_t* p_data, uint64_t size) {
	FILE* p_file;

	if (!p_file_path || !p_data) {
		return -1;
	}

	p_file = fopen(p_file_path, "wb");
	if (!p_file) {
		return -1;
	}

	if (fwrite(p_data, 1, size, p_file) != size) {
		fclose(p_file);
		return -1;
	}

	fclose(p_file);

	return 0;
}

int mmap_file(const char* p_file_path, uint8_t** pp_data, uint64_t* p_size) {
	int fd;
	struct _stat64 stat_buf;
	void* p_data;

	if (!p_file_path || !pp_data || !p_size) {
		return -1;
	}

	fd = open(p_file_path, O_RDONLY);
	if (fd < 0) {
		return -1;
	}

	if (_fstat64(fd, &stat_buf) != 0) {
		close(fd);
		return -1;
	}

	p_data = mmap(NULL, stat_buf.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (!p_data) {
		close(fd);
		return -1;
	}

	close(fd);

	*pp_data = (uint8_t*)p_data;
	*p_size = stat_buf.st_size;

	return 0;
}

int unmmap_file(uint8_t* p_data, uint64_t size) {
	if (!p_data || !size) {
		return -1;
	}

	if (munmap(p_data, size) < 0) {
		return -1;
	}

	return 0;
}

int hmac_sha1(const uint8_t* p_data, uint64_t size, const uint8_t* p_key, uint32_t key_length, uint8_t output[20]) {
	sha1_context context;

	if (!key_length || !output) {
		return -1;
	}

	memset(&context, 0, sizeof(sha1_context));

	sha1_hmac_starts(&context, p_key, key_length);
	sha1_hmac_update(&context, p_data, size);
	sha1_hmac_finish(&context, output);

	memset(&context, 0, sizeof(sha1_context));

	return 0;
}

